@@ -34,6 +34,7 @@ import json
 import shutil
 import hashlib
 import inspect
+import logging
 import tempfile
 import warnings
 import contextlib
@@ -54,27 +55,6 @@ SHAPE_MISMATCH_ERROR = """Error: Image dimensions did not match.
     {actual_path}"""
 
 
-def _download_file(baseline, filename):
-    # Note that baseline can be a comma-separated list of URLs that we can
-    # then treat as mirrors
-    for base_url in baseline.split(','):
-        try:
-            u = urlopen(base_url + filename)
-            content = u.read()
-        except Exception as e:
-            warnings.warn('Downloading {0} failed: {1}'.format(base_url + filename, e))
-        else:
-            break
-    else:
-        raise Exception("Could not download baseline image from any of the "
-                        "available URLs")
-    result_dir = Path(tempfile.mkdtemp())
-    filename = result_dir / 'downloaded'
-    with open(str(filename), 'wb') as tmpfile:
-        tmpfile.write(content)
-    return Path(filename)
-
-
 def _hash_file(in_stream):
     """
     Hashes an already opened file.
@@ -181,8 +161,6 @@ def pytest_configure(config):
         if generate_dir is not None:
             if baseline_dir is not None:
                 warnings.warn("Ignoring --mpl-baseline-path since --mpl-generate-path is set")
-            if results_dir is not None and generate_dir is not None:
-                warnings.warn("Ignoring --mpl-result-path since --mpl-generate-path is set")
 
         if baseline_dir is not None and not baseline_dir.startswith(("https", "http")):
             baseline_dir = os.path.abspath(baseline_dir)
@@ -283,11 +261,23 @@ class ImageComparison:
             self.results_dir = Path(tempfile.mkdtemp(dir=self.results_dir))
         self.results_dir.mkdir(parents=True, exist_ok=True)
 
+        # Decide what to call the downloadable results hash library
+        if self.hash_library is not None:
+            self.results_hash_library_name = self.hash_library.name
+        else:  # Use the first filename encountered in a `hash_library=` kwarg
+            self.results_hash_library_name = None
+
         # We need global state to store all the hashes generated over the run
         self._generated_hash_library = {}
         self._test_results = {}
         self._test_stats = None
 
+        # https://stackoverflow.com/questions/51737378/how-should-i-log-in-my-pytest-plugin
+        # turn debug prints on only if "-vv" or more passed
+        level = logging.DEBUG if config.option.verbose > 1 else logging.INFO
+        logging.basicConfig(level=level)
+        self.logger = logging.getLogger('pytest-mpl')
+
     def get_compare(self, item):
         """
         Return the mpl_image_compare marker for the given item.
@@ -360,6 +350,26 @@ class ImageComparison:
 
         return baseline_dir
 
+    def _download_file(self, baseline, filename):
+        # Note that baseline can be a comma-separated list of URLs that we can
+        # then treat as mirrors
+        for base_url in baseline.split(','):
+            try:
+                u = urlopen(base_url + filename)
+                content = u.read()
+            except Exception as e:
+                self.logger.info(f'Downloading {base_url + filename} failed: {repr(e)}')
+            else:
+                break
+        else:
+            raise Exception("Could not download baseline image from any of the "
+                            "available URLs")
+        result_dir = Path(tempfile.mkdtemp())
+        filename = result_dir / 'downloaded'
+        with open(str(filename), 'wb') as tmpfile:
+            tmpfile.write(content)
+        return Path(filename)
+
     def obtain_baseline_image(self, item, target_dir):
         """
         Copy the baseline image to our working directory.
@@ -374,7 +384,7 @@ class ImageComparison:
         if baseline_remote:
             # baseline_dir can be a list of URLs when remote, so we have to
             # pass base and filename to download
-            baseline_image = _download_file(baseline_dir, filename)
+            baseline_image = self._download_file(baseline_dir, filename)
         else:
             baseline_image = (baseline_dir / filename).absolute()
 
@@ -390,11 +400,14 @@ class ImageComparison:
         if not os.path.exists(self.generate_dir):
             os.makedirs(self.generate_dir)
 
-        fig.savefig(str((self.generate_dir / self.generate_filename(item)).absolute()),
-                    **savefig_kwargs)
+        baseline_filename = self.generate_filename(item)
+        baseline_path = (self.generate_dir / baseline_filename).absolute()
+        fig.savefig(str(baseline_path), **savefig_kwargs)
 
         close_mpl_figure(fig)
 
+        return baseline_path
+
     def generate_image_hash(self, item, fig):
         """
         For a `matplotlib.figure.Figure`, returns the SHA256 hash as a hexadecimal
@@ -435,6 +448,7 @@ class ImageComparison:
 
         if not os.path.exists(baseline_image_ref):
             summary['status'] = 'failed'
+            summary['image_status'] = 'missing'
             error_message = ("Image file not found for comparison test in: \n\t"
                              f"{self.get_baseline_directory(item)}\n"
                              "(This is expected for new tests.)\n"
@@ -456,6 +470,7 @@ class ImageComparison:
         actual_shape = imread(str(test_image)).shape[:2]
         if expected_shape != actual_shape:
             summary['status'] = 'failed'
+            summary['image_status'] = 'diff'
             error_message = SHAPE_MISMATCH_ERROR.format(expected_path=baseline_image,
                                                         expected_shape=expected_shape,
                                                         actual_path=test_image,
@@ -467,10 +482,12 @@ class ImageComparison:
         summary['tolerance'] = tolerance
         if results is None:
             summary['status'] = 'passed'
+            summary['image_status'] = 'match'
             summary['status_msg'] = 'Image comparison passed.'
             return None
         else:
             summary['status'] = 'failed'
+            summary['image_status'] = 'diff'
             summary['rms'] = results['rms']
             diff_image = (result_dir / 'result-failed-diff.png').absolute()
             summary['diff_image'] = diff_image.relative_to(self.results_dir).as_posix()
@@ -496,6 +513,10 @@ class ImageComparison:
         compare = self.get_compare(item)
         savefig_kwargs = compare.kwargs.get('savefig_kwargs', {})
 
+        if not self.results_hash_library_name:
+            # Use hash library name of current test as results hash library name
+            self.results_hash_library_name = Path(compare.kwargs.get("hash_library", "")).name
+
         hash_library_filename = self.hash_library or compare.kwargs.get('hash_library', None)
         hash_library_filename = (Path(item.fspath).parent / hash_library_filename).absolute()
 
@@ -512,14 +533,17 @@ class ImageComparison:
 
         if baseline_hash is None:  # hash-missing
             summary['status'] = 'failed'
+            summary['hash_status'] = 'missing'
             summary['status_msg'] = (f"Hash for test '{hash_name}' not found in {hash_library_filename}. "
                                      f"Generated hash is {test_hash}.")
         elif test_hash == baseline_hash:  # hash-match
             hash_comparison_pass = True
             summary['status'] = 'passed'
+            summary['hash_status'] = 'match'
             summary['status_msg'] = 'Test hash matches baseline hash.'
         else:  # hash-diff
             summary['status'] = 'failed'
+            summary['hash_status'] = 'diff'
             summary['status_msg'] = (f"Hash {test_hash} doesn't match hash "
                                      f"{baseline_hash} in library "
                                      f"{hash_library_filename} for test {hash_name}.")
@@ -544,7 +568,8 @@ class ImageComparison:
             except Exception as baseline_error:  # Append to test error later
                 baseline_comparison = str(baseline_error)
             else:  # Update main summary
-                for k in ['baseline_image', 'diff_image', 'rms', 'tolerance', 'result_image']:
+                for k in ['image_status', 'baseline_image', 'diff_image',
+                          'rms', 'tolerance', 'result_image']:
                     summary[k] = summary[k] or baseline_summary.get(k)
 
             # Append the log from image comparison
@@ -597,9 +622,12 @@ class ImageComparison:
                     remove_ticks_and_titles(fig)
 
                 test_name = self.generate_test_name(item)
+                result_dir = self.make_test_results_dir(item)
 
                 summary = {
                     'status': None,
+                    'image_status': None,
+                    'hash_status': None,
                     'status_msg': None,
                     'baseline_image': None,
                     'diff_image': None,
@@ -614,21 +642,23 @@ class ImageComparison:
                 # reference images or simply running the test.
                 if self.generate_dir is not None:
                     summary['status'] = 'skipped'
+                    summary['image_status'] = 'generated'
                     summary['status_msg'] = 'Skipped test, since generating image.'
-                    self.generate_baseline_image(item, fig)
-                    if self.generate_hash_library is None:
-                        self._test_results[str(pathify(test_name))] = summary
-                        pytest.skip("Skipping test, since generating image.")
+                    generate_image = self.generate_baseline_image(item, fig)
+                    if self.results_always:  # Make baseline image available in HTML
+                        result_image = (result_dir / "baseline.png").absolute()
+                        shutil.copy(generate_image, result_image)
+                        summary['baseline_image'] = \
+                            result_image.relative_to(self.results_dir).as_posix()
 
                 if self.generate_hash_library is not None:
+                    summary['hash_status'] = 'generated'
                     image_hash = self.generate_image_hash(item, fig)
                     self._generated_hash_library[test_name] = image_hash
-                    summary['result_hash'] = image_hash
+                    summary['baseline_hash'] = image_hash
 
                 # Only test figures if not generating images
                 if self.generate_dir is None:
-                    result_dir = self.make_test_results_dir(item)
-
                     # Compare to hash library
                     if self.hash_library or compare.kwargs.get('hash_library', None):
                         msg = self.compare_image_to_hash_library(item, fig, result_dir, summary=summary)
@@ -645,12 +675,15 @@ class ImageComparison:
                             for image_type in ['baseline_image', 'diff_image', 'result_image']:
                                 summary[image_type] = None  # image no longer exists
                     else:
-                        self._test_results[str(pathify(test_name))] = summary
+                        self._test_results[test_name] = summary
                         pytest.fail(msg, pytrace=False)
 
                 close_mpl_figure(fig)
 
-                self._test_results[str(pathify(test_name))] = summary
+                self._test_results[test_name] = summary
+
+                if summary['status'] == 'skipped':
+                    pytest.skip(summary['status_msg'])
 
         if item.cls is not None:
             setattr(item.cls, item.function.__name__, item_function_wrapper)
@@ -667,21 +700,36 @@ class ImageComparison:
         """
         Save out the hash library at the end of the run.
         """
+        result_hash_library = self.results_dir / (self.results_hash_library_name or "temp.json")
         if self.generate_hash_library is not None:
             hash_library_path = Path(config.rootdir) / self.generate_hash_library
             hash_library_path.parent.mkdir(parents=True, exist_ok=True)
             with open(hash_library_path, "w") as fp:
                 json.dump(self._generated_hash_library, fp, indent=2)
+            if self.results_always:  # Make accessible in results directory
+                # Use same name as generated
+                result_hash_library = self.results_dir / hash_library_path.name
+                shutil.copy(hash_library_path, result_hash_library)
+        elif self.results_always and self.results_hash_library_name:
+            result_hashes = {k: v['result_hash'] for k, v in self._test_results.items()
+                             if v['result_hash']}
+            if len(result_hashes) > 0:  # At least one hash comparison test
+                with open(result_hash_library, "w") as fp:
+                    json.dump(result_hashes, fp, indent=2)
 
         if self.generate_summary:
+            kwargs = {}
             if 'json' in self.generate_summary:
                 summary = self.generate_summary_json()
                 print(f"A JSON report can be found at: {summary}")
+            if result_hash_library.exists():  # link to it in the HTML
+                kwargs["hash_library"] = result_hash_library.name
             if 'html' in self.generate_summary:
-                summary = generate_summary_html(self._test_results, self.results_dir)
+                summary = generate_summary_html(self._test_results, self.results_dir, **kwargs)
                 print(f"A summary of the failed tests can be found at: {summary}")
             if 'basic-html' in self.generate_summary:
-                summary = generate_summary_basic_html(self._test_results, self.results_dir)
+                summary = generate_summary_basic_html(self._test_results, self.results_dir,
+                                                      **kwargs)
                 print(f"A summary of the failed tests can be found at: {summary}")
 
 
